#include <stdio.h>
#include <stdlib.h>
#include <xmms/plugin.h>
#include <string.h>
#include <math.h>

#include <xmms/util.h>
#include <pthread.h>

// following also includes some X stuff we need
#include "getroot.c"

// some defines

/* following can be modified at your pleasure */
#define POSX 0
#define POSY 14
#define HEIGHT 64 // 64 maximum height/amplitude of a bar
#define CUTOFF 180 // 180 frequencies are represented by 256 values
// we're cutting off high frequencies by only showing 0 - CUTOFF
#define DIV 1 // we have CUTOFF sources, every bar represents DIV sources
#define SPACE 1 // space between bars
#define WIDTH 1 // width of a bar
#define ORIENTATION 0 // 0 = bottom->up, 1 = top->down, 2 = left->right, 3 = right->left

// set following to HEIGHT to disable falloff
#define FALLOFF 5 // how many pixels the bars should falloff every frame
#define FPS	30 // how many frames per second should be drawn

// color of the bars (0 - 255):
#define RED0 163
#define GREEN0 196
#define BLUE0 34
#define RED1 184
#define GREEN1 221
#define BLUE1 39
#define RED2 205
#define GREEN2 246
#define BLUE2 43
#define RED3 230
#define GREEN3 255
#define BLUE3 100

/* Linearity of the amplitude scale (0.5 for linear, keep in [0.1, 0.9]) */
#define d 0.33

#define DEBUG 0 // print out debug messages

// Forward declarations
static void rootvis_init(void);
static void rootvis_cleanup(void);
static void rootvis_about(void);
static void rootvis_configure(void);
static void rootvis_playback_start(void);
static void rootvis_playback_stop(void);
static void rootvis_render_freq(gint16 freq_data[2][256]);

// Callback functions
VisPlugin rootvis_vtable = {
	0, // Handle, filled in by xmms
	0, // Filename, filled in by xmms

	0,                     // Session ID
	"Root Spectrum Analyzer 0.0.3",  // description

	0, // # of PCM channels for render_pcm()
	1, // # of freq channels wanted for render_freq()    == MONO?

	rootvis_init,           // Called when plugin is enabled
	rootvis_cleanup,        // Called when plugin is disabled
	rootvis_about,          // Show the about box
	rootvis_configure,      // Show the configure box
	0,                     // Called to disable plugin, filled in by xmms
	rootvis_playback_start, // Called when playback starts
	rootvis_playback_stop,  // Called when playback stops
	0,                     // Render the PCM data, must return quickly
	rootvis_render_freq     // Render the freq data, must return quickly
};

// XMMS entry point
VisPlugin *get_vplugin_info(void) {
	return &rootvis_vtable;
}


// Some evil global variables
pthread_t work_thread;
pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
enum {GO, STOP} thread_control;

gint16 shared_freq_data[2][256];

// X related
Display *display;
int screen;
Window rootWin, Parent;
GC gc;
char *display_name = NULL;

// Some helper stuff

void clean_data(void) {
	pthread_mutex_lock(&mutex1);
	memset(shared_freq_data, 0, sizeof(gint16) * 2 * 256);
	pthread_mutex_unlock(&mutex1);
}

void print_status(char msg[]) {
	if (DEBUG == 1) printf(">> rootvis >> %s\n", msg); // for debug purposes, but doesn't tell much anyway
}

void error_exit(char msg[]) {
	printf("*** ERROR (rootvis): %s\n", msg);
	rootvis_vtable.disable_plugin(&rootvis_vtable);
}

void initialize_X(void) {
	// XGCValues values;
	// unsigned long valuemask;
	print_status("Opening X Display");
	display = XOpenDisplay(display_name);
	if (display == NULL) {
		if (display_name == NULL) display_name = getenv("DISPLAY");
		fprintf(stderr, "cannot connect to X server %s\n",
			display_name ? display_name : "(default)");
		error_exit("Connecting to X server failed");
	}
	print_status("Getting screen and window");
	screen = DefaultScreen(display);
	rootWin = ToonGetRootWindow(display, screen, &Parent);

	/* these variables are used to specify various attributes for the GC. */
	/* initial values for the GC. */
	//values.cap_style = CapButt;
	//values.join_style = JoinBevel;
	/* which values in 'values' to check when creating the GC. */
	// valuemask = NULL;
	print_status("Creating Graphical Context");
	gc = XCreateGC(display, rootWin, 0, NULL);
	if (gc < 0) {
		error_exit("Creating Graphical Context failed");
	}
	print_status("Setting Line Attributes");
	XSetLineAttributes(display, gc, 1, LineSolid, CapButt, JoinBevel);
}

long get_color(char red, char green, char blue) {
	Colormap def_colormap;
 	XColor dacolor;
	dacolor.red = red * 256;
	dacolor.green = green * 256;
	dacolor.blue = blue * 256;
	print_status("Getting Colormap");
	def_colormap = DefaultColormap(display, DefaultScreen(display));
	print_status("Allocationg color");
	if (XAllocColor(display, def_colormap, &dacolor) == 0) {
		error_exit("Color allocation failed");
	}
	return dacolor.pixel;
}

void draw_bar(int i, int color, unsigned short level, unsigned short oldlevel) {
	if ((level > oldlevel)&&(level > 0)) {
		XSetForeground(display, gc, color);
		if (ORIENTATION == 0)
			XFillRectangle(display, rootWin, gc, POSX + i*(WIDTH+SPACE), POSY + HEIGHT - level, WIDTH, level);
		else if (ORIENTATION == 1)
			XFillRectangle(display, rootWin, gc, POSX + i*(WIDTH+SPACE), POSY, WIDTH, level);
		else if (ORIENTATION == 2)
			XFillRectangle(display, rootWin, gc, POSX, POSY + i*(WIDTH+SPACE), level, WIDTH);
		else	XFillRectangle(display, rootWin, gc, POSX + HEIGHT - level, POSY + i*(WIDTH+SPACE), level, WIDTH);
	} else if (level < oldlevel) {
		if (ORIENTATION == 0)
			XClearArea(display, rootWin, POSX + i*(WIDTH+SPACE), POSY + HEIGHT - oldlevel, WIDTH, oldlevel - level, False);
		else if (ORIENTATION == 1)
			XClearArea(display, rootWin, POSX + i*(WIDTH+SPACE), POSY + level, WIDTH, oldlevel - level, False);
		else if (ORIENTATION == 2)
			XClearArea(display, rootWin, POSX + level, POSY + i*(WIDTH+SPACE), oldlevel - level, WIDTH, False);
		else	XClearArea(display, rootWin, POSX + HEIGHT - oldlevel, POSY + i*(WIDTH+SPACE), oldlevel - level, WIDTH, False);

	}
}

// Our worker thread

void* worker_func() {
	long pixel[4];
	gint16 freq_data[2][256];
	static double scale, x00, y00;
	unsigned int i, j, level; // , step = 0;
	unsigned short *level1, *level2, *levelsw;
	
	level1 = (unsigned short*)calloc(256, sizeof(short)); // need to be zeroed out
	level2 = (unsigned short*)malloc(256*sizeof(short));
	if ((level1 == NULL)||(level2 == NULL)) error_exit("Allocation of memory failed");
	
	initialize_X();
	pixel[0] = get_color(RED0, GREEN0, BLUE0);
	pixel[1] = get_color(RED1, GREEN1, BLUE1);
	pixel[2] = get_color(RED2, GREEN2, BLUE2);
	pixel[3] = get_color(RED3, GREEN3, BLUE3);
	print_status("Ready to draw");
	
	scale = HEIGHT / (log((1 - d) / d) * 4);
        x00 = d*d*32768.0/(2 * d - 1);
        y00 = -log(-x00) * scale;

	while (thread_control == GO) {
		pthread_mutex_lock(&mutex1);
		memcpy(&freq_data, &shared_freq_data, sizeof(gint16)*2*256);
		pthread_mutex_unlock(&mutex1);

		/* instead of copying the old level array to the second array, 
		we just tell the first is now the second one */
		levelsw = level1;
		level1 = level2;
		level2 = levelsw;

		/* if (step == 1) {
			XClearArea(display, rootWin, POSX, POSY, (CUTOFF/DIV)*(WIDTH+SPACE), POSY + HEIGHT, False);
			step--;
		} else step++; */
	
		for (i = 0; i < CUTOFF/DIV; i++) {
			level = 0;
			for (j = i*DIV; j < (i+1)*DIV; j++)
				if (level < freq_data[0][j])
					level = freq_data[0][j];
			level = level * (i*DIV + 1);
			level = floor(log(level - x00)*scale + y00);
			if (level < HEIGHT) {
				if ((level2[i] > FALLOFF)&&(level < level2[i] - FALLOFF))
					level1[i] = level2[i] - FALLOFF;
				else	level1[i] = level;
			} else level1[i] = HEIGHT;
			draw_bar(i, pixel[(int)ceil((float)level1[i] / HEIGHT * 4) - 1], level1[i], level2[i]);
		}
		XFlush(display);
		xmms_usleep(1000000 / FPS);
	}
	print_status("Worker thread: Exiting");
	XClearWindow(display, rootWin);
	XCloseDisplay(display); 
	free(level1);
	free(level2);
	return NULL;
}


// da xmms functions

static void rootvis_init(void) {
	int rc1;
	print_status("Initializing");
	
	clean_data();		
	thread_control = GO;
	if ((rc1 = pthread_create(&work_thread, NULL, worker_func, NULL))) {
		fprintf(stderr, "Thread creation failed: %d\n", rc1);
		error_exit("Thread creation failed");
	}
}

static void rootvis_cleanup(void) {
	print_status("Cleanup... ");
	thread_control = STOP;
	pthread_join(work_thread, NULL);
	print_status("Clean Exit");
}

static void rootvis_about(void)
{
	print_status("About");
}

static void rootvis_configure(void)
{
	print_status("Configure");
}

static void rootvis_playback_start(void)
{
	print_status("Playback starting");
}

static void rootvis_playback_stop(void)
{
	clean_data();
}

static void rootvis_render_freq(gint16 freq_data[2][256]) {
	int channel, bucket;

	pthread_mutex_lock(&mutex1);
	for (channel = 0; channel < 2; channel++) {
	 for (bucket = 0; bucket < 256; bucket++) {
		shared_freq_data[channel][bucket] = freq_data[channel][bucket];
	 }
	}
	pthread_mutex_unlock(&mutex1);
}
